(*****************************************************************************)
(* The metalevel analysis : constraint generation function, coding of        *)
(* non-empty intersection, and implementation of the attacker                *)
(*                                                                           *)
(*                                 Time-stamp: <31-01-2005 Mikael Buchholtz> *)
(*****************************************************************************)
(* Assumes that all names, variables, crypto-points, and labels only contain *)
(* alphanumeric characters, prime (') or underscore (_) and should not begin *)
(* with an _. If not, the ALFP formula generated cannot be parsed by the     *)
(* Succinct Solver.                                                          *)
(*****************************************************************************)

structure Analysis2 =
struct

val default =
    {withAttacker  = true,
     attackerIndex = "",
     mergeExpressionLabels  = true,
     mergeInputVariables = true}

val version1parameters =
    {withAttacker  = true,
     attackerIndex = "0",
     mergeExpressionLabels  = false,
     mergeInputVariables = false
     }

val version = "2.0"

local
    open MLLysa

    fun toStringParam ({withAttacker,
			attackerIndex,
			mergeExpressionLabels,
			mergeInputVariables}) =
	"withAttacker          = "^Bool.toString(withAttacker)^"\n"^
	"attackerIndex         = \""^attackerIndex^"\"\n"^
	"mergeExpressionLabels = "^Bool.toString(mergeExpressionLabels)^"\n"^
	"mergeInputVariables   = "^Bool.toString(mergeInputVariables)

    (*************************************************************************)
    (*                                                                       *)
    (* Auxiliary functions to generate bits and pieces of an ALPF formula.   *)
    (*                                                                       *)
    (*************************************************************************)

    fun after []      e = []
      | after (l::ls) e = l::e::after ls e


    fun between []      e = []
      | between [t]     e = [t]
      | between (l::ls) e = l::e::between ls e


    fun toSymb (n,i) = n^"_"^(String.concat (between i "_"))


    fun canon(NAME(n,_))  = toSymb(n)
      | canon(NAMEP(n,_)) = "_PLUS_"^toSymb(n)
      | canon(NAMEM(n,_)) = "_MINUS_"^toSymb(n)
      | canon(VAR(x,_))   = toSymb(x)
      | canon(_)          = 
	raise Fail "canon only defined for names and variables"


    fun quantify(l,u,Q,x) =
	let 
	    val s =  Q^" "^x^"_"^(Int.toString l)^". "  
	in
	    if l > u then ""
	    else s^quantify(l+1,u,Q,x)
	end


    fun queryVar(l,u,x) =
	let 
	    val s = x^"_"^(Int.toString l)
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^", "^queryVar(l+1,u,x)
	end


    fun queryRel(l,u,R,x) =
	let 
	    val s = R^"("^x^"_"^(Int.toString l)^")"
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^" & "^queryRel(l+1,u,R,x)
	end


    fun queryRels(l,u,R,x1,x2) =
	let 
	    val s = R^"("^x1^"_"^(Int.toString l)^", "^x2^"_"^(Int.toString l)^")"
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^" & "^queryRels(l+1,u,R,x1,x2)
	end

    
    fun getLab(NAME(_,l))         = toSymb(l)
      | getLab(NAMEP(_,l))        = toSymb(l)
      | getLab(NAMEM(_,l))        = toSymb(l)
      | getLab(VAR(_,l))          = toSymb(l)
      | getLab(SENC(_,_,_,_,_,l)) = toSymb(l)
      | getLab(AENC(_,_,_,_,_,l)) = toSymb(l)


    fun getLabList el =
	String.concat(between (List.map getLab  el) ", ")
     	  

    fun NEIList(el, v, i) =
	let fun NEIify([]   ,v,i) = ["1"]
	      | NEIify(e::el,v,i) = 
		let val n = "NEI("^v^"_"^(Int.toString(i))^", "^(getLab(e))^")"
		in n::NEIify(el,v,i+1)
		end
	in
	    String.concat(between (NEIify(el,v,i)) " & ")
	end


    fun cryptoList(cl, v) =
	String.concat(List.map (fn c => "D("^toSymb(v)^", "^toSymb(c)^") & ")  cl)


    fun copyGAMMA(l,u,l1,l2) =
	let val s = "(A U. GAMMA("^l1^"_"^Int.toString(l)^", U) =>"^" GAMMA("^l2^", U))"
	in
	    if u = 0 then " 1 "
	    else if l>u then ""
	    else if l=u then s
	    else s^" & "^copyGAMMA(l+1,u,l1,l2)
	end

    fun querySameVar(0,v) = ""
      | querySameVar(1,v) = v
      | querySameVar(k,v) = v^", "^querySameVar(k-1,v)


    (* Functions to manage environment Gamma *)

    fun insert (gamma,e)               = e::gamma

    fun lookup (gamma, SET(s))         = s
      | lookup (gamma, UNION(s1,s2))   = (lookup (gamma, s1))@(lookup (gamma, s2))
      | lookup ((x,y)::gamma, SID(x2)) = if x=SID(x2) then y else lookup (gamma, SID(x2))
      | lookup ([], SID(x2) )          = 
	raise Fail ("Set identified "^x2^" used without being declared in a let-declaration.")


    (* Reachability predicate *)
    val rc = ref 0
    fun newReachCount() = (rc:= !rc+1;"RC"^Int.toString(!rc))


    fun insertRho(xl, v, i) =
	let fun rhoify([]   ,v,i) = []
	      | rhoify(x::xl,v,i) = 
		let val r = "RHO("^canon(x)^", "^v^"_"^Int.toString(i)^")" 
		in r::rhoify(xl,v,i+1) 
		end	
	in     
	    String.concat(after (rhoify(xl, v, i)) " & ")
	end


    (*****************************************************************************)
    (*                                                                           *)
    (* genTerm: MLLysa.Term -> ALFP logic                                        *)
    (* genProc: MLLysa.Proc -> ALFP logic                                        *)
    (*                                                                           *)
    (*****************************************************************************)

    fun genTerm(n as NAME(_,l))  = "GAMMA("^toSymb(l)^", "^canon(n)^")\n"
			   
      | genTerm(n as NAMEP(_,l)) = "GAMMA("^toSymb(l)^", "^canon(n)^")\n"

      | genTerm(n as NAMEM(_,l)) = "GAMMA("^toSymb(l)^", "^canon(n)^")\n"
			  
      | genTerm(x as VAR(_,l))   = "(A L'. RHO("^canon(x)^", L') => "^
				   "(A U.  GAMMA(L', U) => GAMMA("^toSymb(l)^", U)))\n"

      | genTerm(SENC(el,e0,c,cl,ls,l)) 
                                 = genTermList(e0::el)^" & \n"^				      
				   "SE"^Int.toString(List.length(el))^
				   "("^toSymb(l)^", "^getLabList(e0::el)^", "^
				   toSymb(c)^", "^toSymb(ls)^") & \n"^
				   cryptoList(cl,ls)^
				   "GAMMA("^toSymb(l)^", "^toSymb(l)^")"			  
			  
      | genTerm(AENC(el,e0,c,cl,ls,l)) 
	                         = genTermList(e0::el)^" &\n"^
				   "AE"^Int.toString(List.length(el))^
				   "("^toSymb(l)^", "^getLabList(e0::el)^", "^
				   toSymb(c)^", "^toSymb(ls)^") & \n"^
				   cryptoList(cl,ls)^
				   "GAMMA("^toSymb(l)^", "^toSymb(l)^")"

    and genTermList el           = String.concat(between ("1"::(List.map genTerm el)) " & " )


    fun genProc gamma b (LET(x,s,p)) = 
	genProc (insert(gamma,(x,lookup(gamma,s)))) b p 

      | genProc gamma b (IPAR(i,s,p)) = 
	"\n/* INDEXED PARALLEL */\n"^
	String.concat
	    ((List.map 
		  (fn e => "("^(genProc gamma b (substIndexProc (i,e) b p))^")&\n")
		  (lookup(gamma,s)))@["1"])

      | genProc gamma b (INEW(al,n,p)) = "\n/* INDEXED RESTRICTION */\n"^
					 genProc gamma b p

      | genProc gamma b (IANEW(al,n,p))= "\n/* INDEXED ASYMMETRIC RESTRICTION */\n"^
					 genProc gamma b p
      
      | genProc gamma b (OUT(el,p)) = 
	let val k = List.length el 
	in
	    "\n/* OUTPUT -length "^Int.toString(k)^" */\n"^
	    genTermList el ^ " & \n" ^ 
	    "KAPPA"^(Int.toString(k))^"("^getLabList el^") & \n" ^
	    "("^(genProc gamma b p)^")"
	end


      (*****************************************************************************)
      (* Below is an analysis with an explicit reachability component rather than  *)
      (* just using implication (=>). Works more efficiently when input/decryption *)
      (* matches lots of values as in version 1.                                   *)
      (*****************************************************************************)

      | genProc gamma b (INP(el,xl,p)) = 
	let val j  = List.length el 
	    val k  = j + List.length xl 
	    val lp = "L'"
	    val rc =  newReachCount()
	in 
	    "\n/* INPUT - length "^Int.toString(k)^" */"^
	    genTermList el ^ " & \n" ^
	    "("^quantify(1,k,"A",lp)^"\n"^
	    "KAPPA"^(Int.toString(k))^"("^queryVar(1,k,lp)^") =>\n"^
	    "(("^NEIList(el,lp,1)^") => \n" ^ 
	    insertRho(xl,lp,j+1)^
	    "REACH("^rc^"))) & \n"^
	    "(REACH("^rc^") => "^(genProc gamma  b p)^")"				
	end

      | genProc gamma b (SDEC(e,el,xl,e0,c,cl,ls,p)) = 
	let val j  = (List.length el)
	    val k  = (List.length xl)+j
	    val lp = "L'"
	    val rc =  newReachCount()
		      
	in
	    "\n/* DECRYPTION - length "^Int.toString(k)^" */\n"^
	    genTerm(e)^" & \n"^
	    cryptoList(cl,ls)^"\n"^
	    genTerm(e0)^" & "^
	    genTermList(el)^" & \n"^
	    "(A L. "^quantify(0,k,"A",lp)^"A ELL0. A L0S. A L'. \n"^
	    "SE"^Int.toString(k)^"(L',"^queryVar(0,k,lp)^", ELL0, L0S) & "^
	    "GAMMA("^getLab(e)^", L') => \n"^
	    "(("^NEIList((e0::el),lp,0)^") => \n" ^ 
	    insertRho(xl,lp,j+1)^
	    "PQD(L0S, "^toSymb(c)^", "^toSymb(ls)^", ELL0) & \n"^
	    "REACH("^rc^"))) &\n"^
	    "(REACH("^rc^") => "^(genProc gamma b p)^")"	
	end

      | genProc gamma b (ADEC(e,el,xl,e0,c,cl,ls,p)) = 
	let val j  = (List.length el)
	    val k  = (List.length xl)+j
	    val lp = "L'"
	    val rc =  newReachCount()		      
	in
	    "\n/* ASYMMETRIC DECRYPTION - length "^Int.toString(k)^" */\n"^
	    genTerm(e)^" &\n"^
	    cryptoList(cl,ls)^"\n"^
	    genTermList(e0::el)^" &\n"^
	    "(A "^lp^". "^quantify(0,k,"A",lp)^" A ELL'. A LS'. \n"^
	    "AE"^Int.toString(k)^"("^lp^", "^queryVar(0,k,lp)^", ELL', LS') & "^
	    "GAMMA("^getLab(e)^", "^lp^") =>\n"^
	    "(E V0. E V0'. GAMMA("^getLab(e0)^", V0) & GAMMA("^lp^"_0, V0') & KP(V0,V0') &"^
	    NEIList(el,lp,1)^") =>\n"^
	    insertRho(xl,lp,j+1)^
	    "PQD(LS', "^toSymb(c)^", "^toSymb(ls)^", ELL') & \n"^
	    "REACH("^rc^")) &\n"^
	    "(REACH("^rc^") => "^(genProc gamma b p)^")"
	end

      | genProc gamma b (NEW(n,p))   = 
	"\n/* RESTRICTION */\n"^"("^(genProc gamma b p)^")"

      | genProc gamma b (ANEW(n,p))  = 
	"\n/* KEY PAIR RESTRICTION */\n"^"("^(genProc gamma b p)^")"

      | genProc gamma b (BANG(p))    = 
	"\n/* REPLICATION */\n"^"("^(genProc gamma b p)^")"

      | genProc gamma b (PAR(p1,p2)) = 
	"\n/* PARALLEL COMPOSITION */\n"^"("^(genProc gamma b p1)^") & ("^(genProc gamma b p2)^")"

      | genProc gamma b (NIL)        = 
	"\n/* NIL */\n"^"1"


    (*****************************************************************************)
    (*                                                                           *)
    (* Functions to generate auxiliary components for the analysis               *)
    (*                                                                           *)
    (* genNEI : Proc -> ALFP logic                                               *)
    (*       Generates the predicate NEI of all pairs of labels, which represent *)
    (*       languages with non-empty intersections.                             *)
    (*                                                                           *)
    (* includes the function                                                     *)
    (*                                                                           *)
    (* genNames : Proc -> ALFP logic                                             *)
    (*       Generates the predicate N containing all names                      *)
    (*                                                                           *)
    (*****************************************************************************)

    fun genNEI p =
	let val seal = sencArityProc(p)
	    val aeal = aencArityProc(p)
		   
	    fun genSenc k = 
		"\n(A L1. A L2.  \n\t("^quantify(0,k,"E","L1")^"E ELL1. E LS1. E L1'.\n\t "^
 		quantify(0,k,"E","L2")^"E ELL2. E LS2. E L2'.\n\t\t"^		       
		"SE"^(Int.toString(k))^"(L1',"^queryVar(0,k,"L1")^", ELL1, LS1) &\n\t\t" ^ 
		"SE"^(Int.toString(k))^"(L2',"^queryVar(0,k,"L2")^", ELL2, LS2) &\n\t\t"^
		queryRels(0,k,"NEI","L1","L2")^"&\n\t\t"^ 
		"GAMMA(L1, L1') &\n\t\t"^
		"GAMMA(L2, L2')"^
		") =>\n\t\t\tNEI(L1,L2))\n"

	    fun genAenc k = 
		"\n(A L1. A L2.  \n\t("^quantify(0,k,"E","L1")^"E ELL1. E LS1. E L1'.\n\t "^
 		quantify(0,k,"E","L2")^"E ELL2. E LS2. E L2'.\n\t\t"^		       
		"AE"^(Int.toString(k))^"(L1',"^queryVar(0,k,"L1")^", ELL1, LS1) &\n\t\t" ^ 
		"AE"^(Int.toString(k))^"(L2',"^queryVar(0,k,"L2")^", ELL2, LS2) &\n\t\t"^
		queryRels(0,k,"NEI","L1","L2")^"&\n\t\t"^ 
		"GAMMA(L1, L1') &\n\t\t"^       
		"GAMMA(L2, L2')"^      
		") =>\n\t\t\tNEI(L1,L2))\n"
	in
	    "(A L1. A L2.\n\t(E U. N(U) & GAMMA(L1,U)  & GAMMA(L2,U))\n\t\t=> NEI(L1,L2))"^
	    (String.concat(map (fn x => "&\n"^genSenc(x)) seal))^
	    (String.concat(map (fn x => "&\n"^genAenc(x)) aeal))
	end
	

    (*****************************************************************************)
    (* Values used by the attacker                                               *)
    (*****************************************************************************)

    val LDY  = ("_L_DY",[])
    val LSDY = ("_L_C",[])

    val NDY  = NAME(("_NDY",[]),LDY)
    val MPDY = NAMEP(("_MDY",[]),LDY)
    val MMDY = NAMEM(("_MDY",[]),LDY)
    val XDY  = VAR(("_XDY",[]),LDY)


    (*****************************************************************************)
    (* Hardest attacker                                                          *)
    (*****************************************************************************)

    fun Phard(p) =
	let
	    fun parList([])    = NIL
	      | parList([p])   = p
	      | parList(p::pl) = List.foldl (fn(p1,p2) => PAR(p2,p1)) p pl

	    fun xdyList(0) = []
	      | xdyList(k) = XDY::xdyList(k-1)

	    fun ldyName(NAME(n,_))  = NAME(n,LDY)
	      | ldyName(NAMEP(n,_)) = NAMEP(n,LDY)
	      | ldyName(NAMEM(n,_)) = NAMEM(n,LDY)
	      | ldyName(_)          = raise Fail "Name expected"

    
	    fun h1(N) = PAR(INP([],[XDY],
			    OUT([XDY],NIL)),
			        parList(List.map (fn n => OUT([ldyName(n)], NIL)) 
						 (NDY::MPDY::MMDY::N)))

	    fun h2(k) = INP([],[XDY],
			OUT(xdyList(k),
			INP([],xdyList(k),NIL)))

	    fun h3(k) = INP([],[XDY],
			OUT([SENC(xdyList(k),XDY,CPDY,[],LSDY,LDY)],
			SDEC(XDY,[],xdyList(k),XDY,CPDY,[],LSDY,NIL)))

	    fun h4(k) = INP([],[XDY],
		        OUT([AENC(xdyList(k),XDY,CPDY,[],LSDY,LDY)],
	                ADEC(XDY,[],xdyList(k),XDY,CPDY,[],LSDY,NIL)))
	in
	    parList((h1 (mfnProc p))::
		    (List.map h2 (comArityProc p) )@
		    (List.map h3 (sencArityProc p))@
		    (List.map h4 (aencArityProc p)))
	end


    (*****************************************************************************)
    (* relabel  g {el,iv} Term/Proc :                                            *)
    (* el : bool   when el is true then labels of names and variables are merged *)
    (*             according to their canonical names                            *)
    (* il : bool   when il is true then variables bound by input are replaced by *)
    (*             XDY                                                           *)
    (*****************************************************************************)

    fun base(VAR(x,l)) = x
      | base(_)        = raise Fail "VAR expected"

    fun newLabel(SENC(tl,t0,c,dl,ls,l)) = l				      
      | newLabel(AENC(tl,t0,c,dl,ls,l)) = l
      | newLabel(NAME((n,il),_)) = ("L"^n,il)
      | newLabel(NAMEP((n,il),_)) = ("L_PLUS_"^n,il)
      | newLabel(NAMEM((n,il),_)) = ("L_MINUS_"^n,il)
      | newLabel(VAR((x,il),_))   = ("L"^x,il)
				      
    fun relabelTerm g {el,iv} (na as NAME(n,l))  = if el 
						   then NAME(n,newLabel(na))
						   else na
      | relabelTerm g {el,iv} (na as NAMEP(n,l)) = if el 
						   then NAMEP(n,newLabel(na))
						   else na
      | relabelTerm g {el,iv} (na as NAMEM(n,l)) = if el 
						   then NAMEM(n,newLabel(na))
						   else na
      | relabelTerm g {el,iv} (na as VAR(n,l))   = if iv andalso List.exists (fn x => x=n) g
						   then XDY
						   else if el
					                then VAR(n,newLabel(na))
					                else na
							     
      | relabelTerm g p (SENC(tl,t0,c,dl,ls,l)) =
	SENC(List.map (relabelTerm g p)  tl,
	     relabelTerm g p t0,
	     c,dl,ls,l)
	
      | relabelTerm g p (AENC(tl,t0,c,dl,ls,l)) =
	AENC(List.map (relabelTerm g p)  tl,
	     relabelTerm g p t0,
	     c,dl,ls,l)

in
    fun relabelProc g p (LET(x,s,P))   = LET(x,s,relabelProc g p P)
      | relabelProc g p (IPAR(i,s,P))  = IPAR(i,s,relabelProc g p P)
      | relabelProc g p (INEW(i,n,P))  = INEW(i,n,relabelProc g p P)
      | relabelProc g p (IANEW(i,n,P)) = IANEW(i,n,relabelProc g p P)
      | relabelProc g p (OUT(tl,P))    = 
	OUT(List.map (relabelTerm g p) tl,
	    relabelProc g p P)
      | relabelProc g p (INP(tl,xl,P)) = 
	let val g' = (Set.union(List.map base xl,g))
	in
	    INP(List.map (relabelTerm g p) tl,
		List.map (relabelTerm g' p) xl,
		relabelProc g' p P)
	end
      | relabelProc g p (SDEC(t,tl,xl,t0,l,ol,ls,P)) =
	SDEC(relabelTerm g p t,
	     List.map (relabelTerm g p) tl,
	     List.map (relabelTerm g p) xl,
	     relabelTerm g p t0,
	     l,ol,ls,
	     relabelProc g p P)
      | relabelProc g p (ADEC(t,tl,xl,t0,l,ol,ls,P)) =
	ADEC(relabelTerm g p t,
	     List.map (relabelTerm g p) tl,
	     List.map (relabelTerm g p) xl,
	     relabelTerm g p t0,
	     l,ol,ls,
	     relabelProc g p P)
	
      | relabelProc g p (NEW(n,P))    = 
	NEW(relabelTerm g p n, 
	    relabelProc g p P)
      | relabelProc g p (ANEW(n,P))   = 
	ANEW(relabelTerm g p n, 				     
	     relabelProc g p P)
      | relabelProc g p (BANG(P))     =  BANG(relabelProc g p P)
      | relabelProc g p (PAR(P1,P2))  =  PAR(relabelProc g p P1,
					     relabelProc g p P2)
      | relabelProc g p (NIL)         = NIL


(*****************************************************************************)
(* generate: MLLysa.Proc -> ALFP logic                                       *)
(*                                                                           *)
(* Main fuction: generates the formula for the analysis of a process         *)
(*****************************************************************************)

fun generate (param as 
             {withAttacker, 
	      attackerIndex, 
	      mergeExpressionLabels, 
	      mergeInputVariables}) proc = 

    let fun genNames(p) = 
	    String.concat (List.map 
			       (fn (n as NAME(_,_))  => "N("^canon(n)^") & \n"
				 | (n as NAMEP(m,_)) => 
				   "N("^canon(n)^") & "^
				   "KP("^canon(NAMEP(m,NOLAB))^", "^canon(NAMEM(m,NOLAB))^") & "^
				   "KP("^canon(NAMEM(m,NOLAB))^", "^canon(NAMEP(m,NOLAB))^") &\n"
				 | (n as NAMEM(m,_)) => 
				   "N("^canon(n)^") & "^
				   "KP("^canon(NAMEP(m,NOLAB))^", "^canon(NAMEM(m,NOLAB))^") & "^
				   "KP("^canon(NAMEM(m,NOLAB))^", "^canon(NAMEP(m,NOLAB))^") &\n"
							
				 | _ => raise Fail "Name expected")
			       (namesProc(p)))
					

	val proc = if not(mergeExpressionLabels) andalso not(mergeInputVariables)
		   then proc
		   else relabelProc [] {el=mergeExpressionLabels, iv=mergeInputVariables} proc

	val proc = if withAttacker 
		   then PAR(proc,Phard(proc))
		   else proc
    in

    "(\n"^
    "/****************************************************************/\n"^
    "/* This file is generated with                                  */\n"^
    "/* Lysatool version "^version^"      */\n"^
    "/* with the following parameters:\n"^
    toStringParam(param)^"\n"^
    "                                                                */\n"^
    "/****************************************************************/\n\n"^


    "/****************************************************************/\n"^
    "/* Names and Key Pairs                                          */\n"^
    "/****************************************************************/\n\n"^

    genNames(proc)^"\n\n"^


    "/****************************************************************/\n"^
    "/* Non-empty intersection                                       */\n"^
    "/****************************************************************/\n\n"^

    "(\n"^genNEI(proc)^") &\n\n"^


    "/****************************************************************/\n"^
    "/* Clauses for the process                                      */\n"^
    "/****************************************************************/\n\n"^

    "(\n"^(genProc [] attackerIndex proc)^") &\n\n"^

    
    "/****************************************************************/\n"^
    "/* Set of all crypto points                                     */\n"^
    "/****************************************************************/\n\n"^
    
    cryptoList(Set.union([CPDY], cryptoProc(proc)),LSDY)^"\n"^

    "/****************************************************************/\n"^
    "/* Pending Queries to PSI                                       */\n"^
    "/****************************************************************/\n\n"^

    "1) & \n"^
    "(A L0S. A ELL. A LS. A ELL0. PQD(L0S, ELL, LS, ELL0) => "^
    "((!D(L0S, ELL) | !D(LS, ELL0)) => PSI(ELL0, ELL)))"

    end


fun analyse proc = generate version1parameters proc

end
end;