(*****************************************************************************)
(* LysaASCIIIO                                                               *)
(*                                                                           *)
(* A structure for parsing and outputting LySa to and from ASCII format      *)
(*                                                                           *)
(*                                 Time-stamp: <27-01-2005 Mikael Buchholtz> *)
(*****************************************************************************)

structure LysaASCIIIO =
struct

exception parseError of string

local

    (*************************************************************************)
    (* Declarations for parsing                                              *)
    (*************************************************************************)

    open MLLysa

    structure LysaLrVals = 
    LysaLrValsFun(structure Token = LrParser.Token)

    structure LysaLexer =
    LysaLexerFun(structure Tokens = LysaLrVals.Tokens)

    structure LysaParser = Join(structure ParserData = LysaLrVals.ParserData
				structure Lex        = LysaLexer
				structure LrParser   = LrParser)

    fun noLab (VAR(x,l)) = x
      | noLab (_)        = raise parseError("only variables should occur here")
     

    fun toVar gamma (NAME(n,l)) = if Set.isMember(n,gamma) 
				then VAR(n,l) else NAME(n,l)
      | toVar gamma (SENC(ts,t,c,cl,l1,l2)) 
                              = SENC(List.map (toVar gamma) ts,(toVar gamma) t,c,cl,l1,l2)
      | toVar gamma (AENC(ts,t,c,cl,l1,l2)) 
                              = AENC(List.map (toVar gamma) ts,(toVar gamma) t,c,cl,l1,l2)
      | toVar gamma (t)       = t


    fun bindVar gamma (LET(x,s,P))     = LET(x,s,bindVar gamma P)
      | bindVar gamma (IPAR(i,s,P))    = IPAR(i,s,bindVar gamma P)
      | bindVar gamma (INEW(al,n,P))   = INEW(al,n,bindVar gamma P)
      | bindVar gamma (IANEW(al,n,P))  = IANEW(al,n,bindVar gamma P)
      | bindVar gamma (OUT(tl,P))      = OUT(List.map (toVar gamma) tl,bindVar gamma P)
      | bindVar gamma (INP(ts,xs,P))   = INP(List.map (toVar gamma) ts,
					     xs,bindVar (Set.union(gamma,List.map noLab xs)) P)
      | bindVar gamma (SDEC(t,ts,xs,tk,c,cl,l,P)) 
                                       = SDEC((toVar gamma) t,List.map (toVar gamma) ts, xs,
					     (toVar gamma) tk,c,cl,l,
					     bindVar (Set.union(gamma,List.map noLab xs)) P)
      | bindVar gamma (ADEC(t,ts,xs,tk,c,cl,l,P)) 
                                       = ADEC((toVar gamma) t,List.map (toVar gamma) ts, xs,
					      (toVar gamma) tk,c,cl,l,
					      bindVar (Set.union(gamma,List.map noLab xs)) P)                                 
      | bindVar gamma (NEW(n,P))       = NEW(n,bindVar gamma P)
      | bindVar gamma (ANEW(n,P))      = ANEW(n,bindVar gamma P)
      | bindVar gamma (BANG P)         = BANG(bindVar gamma P)
      | bindVar gamma (PAR(P1,P2))     = PAR(bindVar gamma P1, bindVar gamma P2)
      | bindVar gamma (NIL)            = NIL
in


fun parseStream stream =
    let val lexer    = LysaParser.makeLexer (fn _ => TextIO.inputLine stream)
	val dummyEOF = LysaLrVals.Tokens.EOF(0,0)
		       
	fun errorMessage(s,i,j) = 
	    raise parseError ("line "^(Int.toString i)^": "^s)
	
	val (result, lexer)    = LysaParser.parse(0,lexer, errorMessage,())
	val (nextToken, lexer) = LysaParser.Stream.get lexer
    in
	if LysaParser.sameToken(nextToken,dummyEOF) 
	then (TextIO.closeIn(stream);bindVar [] result)
	else raise parseError "parsing ended before end-of-file"		   
    end
	handle LysaLexer.UserDeclarations.lexError s => 
	       (TextIO.closeIn stream;
		raise parseError s)
end

fun parseFile filename = parseStream (TextIO.openIn filename)

fun parseString str    = parseStream (TextIO.openString str)



local

    (*************************************************************************)
    (* Declarations for ASCII output                                         *)
    (*************************************************************************)
    
    open MLLysa

    fun between []      e = ""
      | between [t]     e = t
      | between (l::ls) e = l^e^(between ls e)


    fun toStringSymb(n,[])           = if (n,[]) = CPDY
				       then "CPDY" else n
      | toStringSymb(n,tl)           = n^"_{"^(between tl ",")^"}"
					     
    fun toStringSymbList([])         = ""
      | toStringSymbList([s])        = toStringSymb(s)
      | toStringSymbList(s::sl)      = toStringSymb(s)^","^toStringSymbList(sl)

    fun toStringISet(SID(x))         = x
      | toStringISet(s as SET(sl))   = if s = NATURAL1  then "NATURAL1" else
                         	       if s = NATURAL2  then "NATURAL2" else
				       if s = NATURAL3  then "NATURAL3" else
				       if s = NATURAL01 then "NATURAL01" else
				       if s = NATURAL02 then "NATURAL02" else
				       if s = NATURAL03 then "NATURAL03" else
				       if s = ZERO      then "ZERO" else	       
				       "{ "^(between sl ",")^" }"

      | toStringISet(UNION(s1,s2))   = toStringISet(s1)^" union "^toStringISet(s2)
				    
    fun toStringTerm(NAME(n,_))      = toStringSymb(n)
      | toStringTerm(NAMEP((n,l),_)) = toStringSymb((n^"+",l))
      | toStringTerm(NAMEM((n,l),_)) = toStringSymb((n^"-",l))
      | toStringTerm(VAR(x,_))       = toStringSymb(x)
      | toStringTerm(SENC(tl,t,c,cl,_,_))  = "{ "^toStringTermList(tl)^" } : "^
				       toStringTerm(t)^
				       " [at "^toStringSymb(c)^
				       " dest {"^toStringSymbList(cl)^"}]"
      | toStringTerm(AENC(tl,t,c,cl,_,_)) = "{| "^toStringTermList(tl)^" |} : "^
				       toStringTerm(t)^
				       " [at "^toStringSymb(c)^
				       " dest {"^toStringSymbList(cl)^"}]"

    and toStringTermList([])         = ""
      | toStringTermList([t])        = toStringTerm(t)
      | toStringTermList(t::tl)      = toStringTerm(t)^", "^toStringTermList(tl)


    fun toStringProc(LET(x,s,P))     = "let "^toStringISet(x)^
				       " subset "^toStringISet(s)^
				       " in\n"^toStringProc(P)

      | toStringProc(IPAR(i,s,P))    = "(|_{"^i^" in "^toStringISet(s)^"} "^
				       toStringProc(P)^")"

      | toStringProc(INEW(al,n,p))   = "(new_{"^
				       (between
					    (List.map
						 (fn (i,s) => i^" in "^toStringISet(s))
						 al)					    
					    ",")^
				       "} "^toStringTerm(n)^")"^toStringProc(p)

      | toStringProc(IANEW(al,n,p))  = "(new_{"^
				       (between
					    (List.map
						 (fn (i,s) => i^" in "^toStringISet(s))
						 al)					    
					    ",")^
				       "} +- "^toStringTerm(n)^")"^toStringProc(p)


      |	toStringProc(OUT(tl,P))      = "<"^toStringTermList(tl)^">.\n"^
				     toStringProc(P)

      | toStringProc(INP(tl,xl,P))   = "("^toStringTermList(tl)^
                                      "; "^toStringTermList(xl)^").\n"^		    
				      toStringProc(P)

      | toStringProc(SDEC(t,tl,xl,t0,c,cl,_,P)) 
                                  = "decrypt "^toStringTerm(t)^" as "^
				    "{ "^ toStringTermList(tl)^
				    "; "^toStringTermList(xl)^
				    " } : "^toStringTerm(t0)^
				    " [at "^toStringSymb(c)^
				    " orig {"^toStringSymbList(cl)^"}]"^
				    " in\n"^toStringProc(P)
      | toStringProc(ADEC(t,tl,xl,t0,c,cl,_,P)) 
	                           = "decrypt "^
				     toStringTerm(t)^" as "^
				     "{| "^toStringTermList(tl)^
				     "; "^toStringTermList(xl)^
				     " |} : "^toStringTerm(t0)^
				     " [at "^toStringSymb(c)^
				     " orig {"^toStringSymbList(cl)^"}]"^
				     "in\n"^toStringProc(P)
      | toStringProc(PAR(P1,P2))      = "("^toStringProc(P1)^"\n|\n"^toStringProc(P2)^")"
      | toStringProc(BANG(P))         = "! "^toStringProc(P)
      | toStringProc(NEW(n,P))        = "(new "^toStringTerm(n)^")(\n"^toStringProc(P)^")"
      | toStringProc(ANEW(n,P))       = "(new +- "^toStringTerm(n)^")(\n"^toStringProc(P)^")"
      | toStringProc(NIL)             = "0"			      


in
    fun toString P = toStringProc(P)

    fun toFile file P = 
	let val fh = TextIO.openOut(file)
	in
	    (TextIO.output(fh, toStringProc(P));
	     TextIO.closeOut(fh))
	end
end
end