(*****************************************************************************)
(* Output functions that produce HTML                                        *)
(*                                                                           *)
(*                                 Time-stamp: <31-01-2005 Mikael Buchholtz> *)
(*                                                                           *)
(* Revised version                                                           *)
(*****************************************************************************)


structure LysaHTMLIO =

struct

local
    open MLLysa
    open TextIO

    val f2name = ref ""

    (*************************************************************************)
    (* Define constants for printing                                         *)
    (*************************************************************************)
    
    val HTMLCPDY = "CPDY"
    val HTMLUCP  = "UCP"
    val HTMLLDY  = "l<sub>&bull;</sub>"
    val HTMLXDY  = "x<sub>&bull;</sub>"
    val HTMLNDY  = "n<sub>&bull;</sub>"
    val HTMLMPDY = "m<sub>&bull;</sub>+"
    val HTMLMMDY = "m<sub>&bull;</sub>->"

    fun toHTMLMap f ([])    = ""
      | toHTMLMap f ([e])   = f(e)
      | toHTMLMap f (e::ls) = f(e)^", "^(toHTMLMap f ls)

    fun toHTMLApp s f ([])    = ()
      | toHTMLApp s f ([e])   = output(s, f(e))
      | toHTMLApp s f (e::ls) = (output(s, f(e)^", ");(toHTMLApp s f ls))


    (*************************************************************************)
    (* HTML output of processes                                              *)
    (*************************************************************************)

    fun toHTMLSymb(n,[])           = if (n,[]) = CPDY then HTMLCPDY else n
      | toHTMLSymb(n,tl)           = n^"_{"^(toHTMLMap (fn i => i) tl)^"}"
	
    fun toHTMLISet(SID(x))         = x
      | toHTMLISet(s as SET(sl))   = if s = NATURAL1  then "NATURAL1" else
				     if s = NATURAL2  then "NATURAL2" else
				     if s = NATURAL3  then "NATURAL3" else
				     if s = NATURAL01 then "NATURAL01" else
				     if s = NATURAL02 then "NATURAL02" else
				     if s = NATURAL03 then "NATURAL03" else
				     if s = ZERO      then "ZERO" else
				     "{ "^(toHTMLMap (fn i => i) sl)^" }"

      | toHTMLISet(UNION(s1,s2))   = toHTMLISet(s1)^" union "^toHTMLISet(s2)


    fun toHTMLRef(n,tl)            = n

    fun toHTMLAnnote(str, c, [])   = 
	if c = ("_UCP_",[]) then ""
	else "[at "^toHTMLSymb(c)^"]"      	
	
      |	toHTMLAnnote(str, c, cl)   = 
	(if c = ("_UCP_",[]) then "["
	 else "[at "^toHTMLSymb(c)^"]")^
	str^" {"^(toHTMLMap toHTMLSymb cl)^"}]"

    fun toHTMLTerm(NAME(n,l))      = 
	"<a href=\""^(!f2name)^"#"^toHTMLRef(l)^"\""^
	"title=\""^toHTMLSymb(l)^"\">"^toHTMLSymb(n)^"</a>"

      | toHTMLTerm(NAMEP((n,il),l)) = 
	"<a href=\""^(!f2name)^"#"^toHTMLRef(l)^"\""^
	"title=\""^toHTMLSymb(l)^"\">"^toHTMLSymb((n^"+",il))^"</a>"

      | toHTMLTerm(NAMEM((n,il),l)) = 
	"<a href=\""^(!f2name)^"#"^toHTMLRef(l)^"\""^
	"title=\""^toHTMLSymb(l)^"\">"^toHTMLSymb(n^"-",il)^"</a>"

      | toHTMLTerm(VAR(x as("_XDY",[]),l))       = 
	"<a href=\""^(!f2name)^"#"^HTMLXDY^
	"\">"^HTMLXDY^"</a>"

      | toHTMLTerm(VAR(x,l))       = 
	"<a href=\""^(!f2name)^"#"^toHTMLRef(x)^
	"\">"^toHTMLSymb(x)^"</a>"

      | toHTMLTerm(SENC(tl,t,c,cl,_,l)) = 
	"{ "^(toHTMLMap toHTMLTerm tl)^
	"<a href=\""^(!f2name)^"#"^toHTMLSymb(l)^"\""^
	"title=\""^toHTMLSymb(l)^"\"> } : </a>"^
	toHTMLTerm(t)^" "^
	toHTMLAnnote("dest", c, cl)	

      | toHTMLTerm(AENC(tl,t,c,cl,_,l)) = 
	"{| "^(toHTMLMap toHTMLTerm tl)^
	"<a href=\""^(!f2name)^"#"^toHTMLRef(l)^"\""^
	"title=\""^toHTMLSymb(l)^"\"> |} : </a>"^	     
	toHTMLTerm(t)^" "^
	toHTMLAnnote("dest", c, cl)
	

    fun toHTMLProc s (LET(x,S,P)) = 
	(output(s, "let "^toHTMLISet(x)^" subset "^toHTMLISet(S)^" in<br>\n");
	 toHTMLProc s P)

      | toHTMLProc s (IPAR(i,S,P)) = 
	(output(s, "(|_{"^i^" in "^toHTMLISet(S)^"} ");
	 toHTMLProc s P;
	 output(s, ")"))

      | toHTMLProc s (INEW(al,n,P)) = 
	(output(s, "(new_{"^
		   (toHTMLMap (fn (i,s) => i^" in "^toHTMLISet(s)) al)^
		   "} "^
		   toHTMLTerm(n)^")");
	 toHTMLProc s P)

      | toHTMLProc s (IANEW(al,n,P)) = 
	(output(s, "(new_{"^
		   (toHTMLMap (fn (i,s) => i^" in "^toHTMLISet(s))  al)^
		   "} +- "^toHTMLTerm(n)^")");
	 toHTMLProc s P)

      |	toHTMLProc s (OUT(tl,P)) =
	(output(s, "&lt;"^(toHTMLMap toHTMLTerm tl)^"&gt;.<br>\n");
	 toHTMLProc s P)

      | toHTMLProc s (INP(tl,xl,P)) = 
	(output(s, "("^(toHTMLMap toHTMLTerm tl)^
                   "; "^(toHTMLMap toHTMLTerm xl)^").<br>\n");
	 toHTMLProc s P)

      | toHTMLProc s (SDEC(t,tl,xl,t0,c,cl,_,P)) =
        (output(s, "decrypt "^toHTMLTerm(t)^" as "^
		   "{ "^(toHTMLMap toHTMLTerm tl)^
		   "; "^(toHTMLMap toHTMLTerm xl)^
		   " } : "^toHTMLTerm(t0)^" "^
		   toHTMLAnnote("orig", c, cl)^
		   " in<br>\n");
	 toHTMLProc s P)

      | toHTMLProc s (ADEC(t,tl,xl,t0,c,cl,_,P)) =
	(output(s, "decrypt "^
		   toHTMLTerm(t)^" as "^
		   "{| "^(toHTMLMap toHTMLTerm tl)^
		   "; "^(toHTMLMap toHTMLTerm xl)^
		   " |} : "^toHTMLTerm(t0)^" "^			     
		   toHTMLAnnote("orig", c, cl)^			     
		   "in<br>\n");
	 toHTMLProc s P)

      | toHTMLProc s (PAR(P1,P2)) = 
	(output(s,"(");
	 toHTMLProc s P1;
	 output(s, "<br>\n|<br>\n");
	 toHTMLProc s P2;
	 output(s, ")"))

      | toHTMLProc s (BANG(P)) = 
	(output(s, "! ");
	 toHTMLProc s P)

      | toHTMLProc s (NEW(n,P)) = 
	(output(s, "(new "^toHTMLTerm(n)^")(<br>\n");
	 toHTMLProc s P;
	 output(s, ")"))

      | toHTMLProc s (ANEW(n,P)) = 
	(output(s,"(new +- "^toHTMLTerm(n)^")(<br>\n");
	 toHTMLProc s P;
	 output(s, ")"))

      | toHTMLProc s (NIL) = 
	output(s, "0")


fun toHTMLFrameProc framename P = 
    let val _ = (f2name:=framename^"_f2.html")	
	val s = openOut (framename^"_f1.html")
    in 
	output(s, "<html>\n");
	output(s, "<head>\n");
	output(s, "<title>LySatool - input process</title>\n");
	output(s, "<style type=\"text/css\">\n");
	output(s, "a {color:black;text-decoration: none}\n");
	output(s, "a:hover {color:red}\n");
	output(s, "</style>\n");
	output(s, "<base href=\""^framename^"_f2.html\" target=\"f2\">\n");    
	output(s, "</head>\n");
	output(s, "<body>\n");    
	toHTMLProc s P;
	output(s, "</body>\n");
	output(s, "</html>\n");
    closeOut s
    end


    (*************************************************************************)
    (* HTML output of analysis results                                       *)
    (*************************************************************************)

    (*************************************************************************)
    (* Functions for manipulation of output relations                        *)
    (*************************************************************************)


    fun getRel(name, [])            = raise Fail ("Relation missing: "^name)
      | getRel(name, (n,rel)::rels) = if name=n then rel else getRel(name,rels)

    
    fun funRel(rel) =
    let fun extend([],            rest, a::el) = (a,[el])::rest
	  | extend((a',el')::res, rest, a::el) = if a = a' 
						 then (a',el::el')::(res@rest)
						 else extend(res, (a',el')::rest, a::el)
	  | extend(_,_,_) = raise Fail "This only happens if extend is called with wrong parameters"

	fun funRel' ([])     res = res
	  | funRel' (el::rl) res = funRel' rl (extend(res, [], el))
    in
	funRel' rel []
    end



    (*************************************************************************)
    (* Functions for layout of elements                                      *)
    (*************************************************************************)

    fun findIndex s = 
	let fun fi([],i)    = i
	      | fi(h::ls,i) = if h = #"_" then i else fi(ls,i+1)
	in 
	    fi(String.explode(s),0)
	end

    fun toHTMLIndices e = (String.translate (fn c => if c = #"_" 
						     then "&nbsp;" 
						     else Char.toString(c))
					    e)

    fun toHTMLElem  e = 
	let val fu   = findIndex e
	    val size = String.size(e) 
	in
	    if fu = size
	    then e
	    else String.substring(e,0,fu)^"<sub>"^
		 toHTMLIndices(String.substring(e,fu+1,size-fu-1))^	
		 "</sub>"
	end
	
    fun toHTMLCP "_CPDY_" = HTMLCPDY
      | toHTMLCP "_UCP__" = HTMLUCP
      | toHTMLCP e        = toHTMLElem e
    
    fun toHTMLVar "_XDY_"  = HTMLXDY
      | toHTMLVar e        = toHTMLElem e

    fun toHTMLName "_NDY_"        = HTMLNDY
      | toHTMLName "_PLUS__MDY_"  = HTMLMPDY
      | toHTMLName "_MINUS__MDY_" = HTMLMMDY
      | toHTMLName e              =
	if String.isPrefix "_PLUS_" e then 
	    toHTMLElem (String.substring(e,6,String.size(e)-6))^"+" else
	if String.isPrefix "_MINUS_" e then 
	    toHTMLElem (String.substring(e,7,String.size(e)-7))^"-" else
	toHTMLElem e

    
    fun toHTMLLabNoRef "_L_DY_" = HTMLLDY
      | toHTMLLabNoRef e        = "L"^(toHTMLName(String.substring(e,1,String.size(e)-1)))

				 
    fun toHTMLLab e             = "<a href=\"#"^e^"\">"^toHTMLLabNoRef(e)^"</a>"


    (*************************************************************************)
    (* Confidential                                                          *)
    (*************************************************************************)

    fun toHTMLConf s rels ({withAttacker,
		      attackerIndex,
		      mergeExpressionLabels,
		      mergeInputVariables}) toHTMLValue =
	if not withAttacker 
	then output(s,"<i>No attacker present</i>\n")
	else let val gamma = funRel(getRel("GAMMA/2",rels))
		 val res   = List.find (fn (x,_) => x="_L_DY_") gamma
	     in
		 case res of
		     SOME((_,ldy)) => toHTMLApp s toHTMLValue (List.concat(ldy))
		   | NONE          => raise Fail ("No attacker present in result")
	     end

    (*************************************************************************)
    (* PSI                                                                   *)
    (*************************************************************************)

    fun toHTMLPSI s rels =
	let val psi = getRel("PSI/2",rels)	      
	in
	    if List.length(psi) = 0 then 
		output(s, "<i>No violations possible</i>") else 
	        toHTMLApp s (fn e => "("^(toHTMLMap toHTMLCP e)^")") psi
	end


    (*************************************************************************)
    (* RHO                                                                   *)
    (*************************************************************************)

    fun toHTMLRHO s rels =
	let val rho = ListMergeSort.sort 
			  (fn ((x,_),(y,_)) => x>y) 
			  (funRel(getRel("RHO/2",rels)))
			  
	    fun toHTMLID "_XDY_" = HTMLXDY
	      | toHTMLID a       = String.substring(a,0,findIndex(a))
				    

	    fun toHTMLEntry s (x,el) = (
		output(s, "<table border=\"0\">\n"^	    
			  "<tr><td valign=\"top\" id=\""^toHTMLID(x)^"\">&rho;("^
			  toHTMLVar(x)^")</td>"^
			  "<td valign=\"top\"> = </td><td valign=\"top\">");
		toHTMLApp s (fn e => toHTMLMap toHTMLLab e)    el;
		output(s, "</td></tr>\n"^
			  "</table>"))
	in
	    List.app (toHTMLEntry s) rho
	end


    (*************************************************************************)
    (* KAPPA                                                                   *)
    (*************************************************************************)

    fun toHTMLKAPPA s rels =

	let fun getKAPPA([])            = []
	      | getKAPPA((n,rel)::rels) = if String.isPrefix "KAPPA" n 
					     then rel::getKAPPA(rels)
					     else getKAPPA(rels)
	    val kappa = List.concat(getKAPPA(rels))
	in	    
	    toHTMLApp s (fn e => "&lt;"^(toHTMLMap toHTMLLab e)^"&gt;")  kappa
	end


    (*************************************************************************)
    (* GAMMA                                                                 *)
    (*************************************************************************)

    fun toHTMLGAMMA s rels toHTMLValue =
	let fun toHTMLID "_L_DY_" = "_L_DY_"
	      | toHTMLID a        = String.substring(a,0,findIndex(a))					       

	    fun toHTMLRule s ((a,el)) = 
		(output(s, "<table border=\"0\">\n"^
			   "<tr id=\""^a^"\">"^
			   "<td id=\""^toHTMLID(a)^"\" valign=\"top\">"^
			   toHTMLLabNoRef(a)^"</td>"^
			   "<td valign=\"top\">&rarr;</td><td valign=\"top\">");
		 (toHTMLApp s toHTMLValue (List.concat el));
		 output(s, "</td></tr>\n"^
			   "</table>\n"))				    
		
	    fun gt((x,_),(y,_)) = let val x = String.substring(x,0,findIndex(x))
				      val y = String.substring(y,0,findIndex(y))
				  in
				      String.size(x) > String.size(y) orelse
				      String.size(x) = String.size(y) andalso
				      String.>(x,y)
				  end
				  
	    val gamma = ListMergeSort.sort gt (funRel(getRel("GAMMA/2",rels)))
			
	in
	    List.app (toHTMLRule s) gamma
	end


    (*************************************************************************)
    (* LEGEND                                                                *)
    (*************************************************************************)

    fun toHTMLParam ({withAttacker,
		      attackerIndex,
		      mergeExpressionLabels,
		      mergeInputVariables}) =
	"<tr><td>withAttacker</td><td>"^Bool.toString(withAttacker)^"</td></tr>\n"^
	"<tr><td>attackerIndex</td><td>\""^attackerIndex^"\"</td></tr>\n"^
	"<tr><td>mergeExpressionLabels</td><td>"^
	Bool.toString(mergeExpressionLabels)^"</td></tr>\n"^
	"<tr><td>mergeInputVariables</td><td>"^
	Bool.toString(mergeInputVariables)^"</td></tr>\n"

    fun toHTMLLegend s param = 
	output(s, "<table border=\"0\" class=\"legend\">\n"^
		  "<tr><td colspan=\"2\">Analysis result produced by</td></tr>\n"^
		  "<tr><td colspan=\"2\">Lysatool version "^Analysis2.version^"\n"^
		  "with the following parameters</td></tr>\n"^
		  toHTMLParam(param)^							   
		  "<tr><td>&nbsp;</td><td>&nbsp;</td></tr>\n"^
		  "<tr><td>"^HTMLNDY^"</td>"^
		  "<td>Name known initially by the attacker</td></tr>\n"^
		  "<tr><td>"^HTMLMPDY^", "^HTMLMMDY^"</td>"^
		  "<td>Key pair known initially by the attacker</td></tr>\n"^
		  "<tr><td>"^HTMLXDY^"</td>"^
		  "<td>Canonical variable of the attacker</td></tr>\n"^
		  "<tr><td>"^HTMLLDY^"</td>"^
		  "<td>Non-terminal in tree grammar used by the attacker</td></tr>\n"^
		  "<tr><td>"^HTMLCPDY^"</td>"^
		  "<td>Crypto-point used at the attacker</td></tr>\n"^
		  "<tr><td>"^HTMLUCP^"</td>"^
		  "<td>Unspecified crypto-point. The pair ("^HTMLUCP^", c) can occur "^
                  "in &psi; if something encrypted at an unannotated encryption may "^
		  "be decrypted at c."^
		  "</td></tr>\n"^ "</table>\n")	


fun toHTMLFrameResult framename param (uni, rels) = 
    let	fun toHTMLValue(v) = 	    
	    let	fun toHTMLDest(d) = 
		    let val res = List.find (fn(x',_) => d=x') (funRel(getRel("D/2",rels)))
		    in 
			case res of 
			    SOME((x,y)) => "{ "^(toHTMLMap toHTMLCP (List.concat(y)))^" }"
			  | _           => raise Fail ("Destination set not in D("^d^")")
		    end

		fun toHTMLAnno("_UCP__","_L_C_") = ""
		  | toHTMLAnno(c, "_L_C_")       = " [at "^toHTMLCP(c)^"]"
		  | toHTMLAnno("_UCP__", d )     = " [dest "^toHTMLDest(d)^"]"
		  | toHTMLAnno(c, d)             = " [at "^toHTMLCP(c)^
						   " dest "^toHTMLDest(d)^"]"	    
	    
		fun toHTMLEnc name sb eb x =
		    let fun getPrefixRel (pre, [])          = []
			  | getPrefixRel (pre, (n,rl)::rel) = if String.isPrefix pre n 
							      then rl@getPrefixRel(pre,rel)
							      else getPrefixRel(pre,rel)
								   
			val frel = funRel(getPrefixRel(name,rels))
				   
			val res = List.find (fn (x',_) => x=x') frel

			fun ps(y) =  let val t0 = List.hd y				 
					 val rev = List.rev(List.tl y)
					 val dl  = List.hd rev
					 val cp  = List.hd (List.tl rev)
					 val el  = List.rev (List.tl (List.tl rev))
				     in				 
					 sb^(toHTMLMap toHTMLLab el)^eb^
					 "<sub>"^toHTMLLab(t0)^"</sub>"^
					 toHTMLAnno(cp,dl)
				     end 
		    in
			case res of			   
			    (SOME((x,y))) => toHTMLMap ps y 
			  | _             => raise (Fail ("function undefined for argument"))
		    end			
	    in		
		toHTMLEnc "SE" "{" "}" v handle Fail(s) =>					 
		    toHTMLEnc "AE" "{|" "|}" v handle Fail(_) =>
			 toHTMLName(v)
	    end


	val s = openOut (framename^"_f2.html")	
    in
	output(s, "<html>\n");
	output(s, "<head>\n");
	output(s, "<title>LySatool - analysis result</title>\n");
	output(s, "<style type=\"text/css\">\n");
	output(s, "a {color:black;text-decoration: none}\n");
	output(s, "a:hover {color:red}\n");
	output(s, "</style>\n");	
	output(s, "</head>\n");
	output(s, "<body>\n");
	output(s, "Values that may not be confidential\n<br>");
	toHTMLConf s rels param toHTMLValue;
	output(s, "<hr>\n\n");
	output(s, "Violation of authentication properties (&psi;)\n<br>");
	toHTMLPSI s rels;
	output(s, "<hr>\n\n");
	output(s, "Variable bindings (&rho;)<br>\n");
	toHTMLRHO s rels;
	output(s, "<hr>\n\n");
	output(s, "Messages communicated (&kappa;)<br>\n");
	toHTMLKAPPA s rels;	
	output(s, "<hr>\n\n");
	output(s, "Tree grammars (&gamma;)<br>\n");
	toHTMLGAMMA s rels toHTMLValue;
	output(s, "<hr>\n\n");
	toHTMLLegend s param;
	output(s, "</body>\n");
	output(s, "</html>\n");
	closeOut s
    end
in

fun toHTML framename P (param as {withAttacker, 
				  attackerIndex, 
				  mergeExpressionLabels, 
				  mergeInputVariables}) result =
    let val s = openOut (framename^".html")
	val _ = toHTMLFrameProc framename 
				(Analysis2.relabelProc [] {el=mergeExpressionLabels,
							   iv=mergeInputVariables} P)

	val _ = toHTMLFrameResult framename param result
	val {dir, file=framename} = OS.Path.splitDirFile(framename)
    in 
    output(s,"<html>\n"^
	     "<head>\n"^
	     "<title>Lysatool ("^Analysis2.version^")</title>\n"^
	     "</head>\n"^
	     "<frameset rows=\"30%,70%\">\n"^
	     "<frame name=\"f1\" src=\""^framename^"_f1.html\">\n"^
	     "<frame name=\"f2\" src=\""^framename^"_f2.html\">\n"^
	     "<frameset>\n"^	
	     "</html>\n");
    closeOut s    
    end
end
end