(*****************************************************************************)
(* LysaLatexIO                                                               *)
(*                                                                           *)
(* Pretty printing in Latex format                                           *)
(*                                                                           *)
(*                                 Time-stamp: <27-01-2005 Mikael Buchholtz> *)
(*****************************************************************************)

structure LysaLatexIO =
struct


local

    (*************************************************************************)
    (* Declarations for Latex output                                         *)
    (*************************************************************************)
    
    open MLLysa

    fun between []      e = ""
      | between [t]     e = t
      | between (l::ls) e = l^e^(between ls e)

    fun toStringNum (n)              = if n >= 0 
				       then Int.toString(n) 
				       else "-"^Int.toString(~1*n)


    fun toStringSymb("_UCP_",[])        = "UCP"
      | toStringSymb("_CPDY",[])        = "\\ell_\\bullet"
      | toStringSymb(n,[])              =  if String.size(n) <= 1 then n else
					"\\mathit{"^n^"}"
      | toStringSymb(n,tl)              =  let val name = if String.size(n) <= 1 then n else
							  "\\mathit{"^n^"}" 
					   in 
					    name^"_{"^(between tl " ")^"}"
					   end
	
    fun toStringSymbPM s (n,[])      = if String.size(n) <= 1 then n^"^"^s else  
				       "\\mathit{"^n^"}^"^s

      | toStringSymbPM s (n,tl)      = if String.size(n) <= 1 
				       then n^"^"^s^"_{"^(between tl " ")^"}"
				       else "\\mathit{"^n^"}^"^s^"_{"^(between tl " ")^"}"
				     
    fun toStringSymbList([])         = ""
      | toStringSymbList([s])        = toStringSymb(s)
      | toStringSymbList(s::sl)      = toStringSymb(s)^", "^toStringSymbList(sl)

    fun toStringISet(SID(x))         = x
      | toStringISet(s as SET(sl))   =
	if s = NATURAL1 then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{1\\}"  else
	if s = NATURAL2  then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{1,2\\}" else
	if s = NATURAL3  then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{1,2,3\\}" else
	if s = NATURAL01 then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{0, 1\\}"  else
	if s = NATURAL02  then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{0, 1,2\\}" else
	if s = NATURAL03  then 
	    "\\mathbb{N}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}} = \\{0, 1,2,3\\}" else
        if s = ZERO      then 
	    "\\mathbb{N}_{-}\\;\\LFONT{s.t.}\\;\\CANON{\\mathbb{N}_{-}} = \\{ 0 \\}" else
	"\\{ "^(between sl ",")^" \\}"
      
      | toStringISet(UNION(s1,s2))   = toStringISet(s1)^"\\cup "^toStringISet(s2)
				    
    fun toStringTerm(NAME(n,_))      = toStringSymb(n)
      | toStringTerm(NAMEP(n,_))     = toStringSymbPM "+" (n)
      | toStringTerm(NAMEM(n,_))     = toStringSymbPM "-" (n)
      | toStringTerm(VAR(x,_))       = toStringSymb(x)
      | toStringTerm(SENC(tl,t,c,cl,_,_))  
                                     = "\\SENC{"^toStringTermList(tl)^"}{"^
				       toStringTerm(t)^"}\\ATDEST{"^
				       toStringSymb(c)^"}{\\{"^
				       toStringSymbList(cl)^"\\}}"
      | toStringTerm(AENC(tl,t,c,cl,_,_)) = 
	                               "\\AENC{"^toStringTermList(tl)^"}{"^
				       toStringTerm(t)^"}\\ATDEST{"^
				       toStringSymb(c)^"}{\\{"^
				       toStringSymbList(cl)^"\\}}"

    and toStringTermList([])         = ""
      | toStringTermList([t])        = toStringTerm(t)
      | toStringTermList(t::tl)      = toStringTerm(t)^", "^toStringTermList(tl)


    fun toStringProc(LET(x,s,P))     = "\\LET{"^toStringISet(x)^"}{"^
				       toStringISet(s)^"}\\IN\\\\\n"^
				       toStringProc(P)

      | toStringProc(IPAR(i,s,P))    = "\\IPAR{"^i^"\\in "^toStringISet(s)^"}\n"^
				       toStringProc(P)

      | toStringProc(INEW(al,n,p))   = "\\INEW{"^
				       (between
					    (List.map
						 (fn (i,s) => i^"\\in "^toStringISet(s))
						 al)					    
					    ", ")^
				       "}{"^toStringTerm(n)^"}\\\\\n"^toStringProc(p)

      | toStringProc(IANEW(al,n,p))  = "\\IANEW{"^
				       (between
					    (List.map
						 (fn (i,s) => i^"\\in "^toStringISet(s))
						 al)					    
					    ", ")^
				       "}{"^toStringTerm(n)^"}\\\\\n"^toStringProc(p)


      |	toStringProc(OUT(tl,P))      = "\\OUT{"^toStringTermList(tl)^"}.\\\\\n"^
				     toStringProc(P)

      | toStringProc(INP(tl,xl,P))   = "\\INP{"^toStringTermList(tl)^"}{"^
				       toStringTermList(xl)^"}.\\\\\n"^		    
				       toStringProc(P)

      | toStringProc(SDEC(t,tl,xl,t0,c,cl,_,P)) 
                                     = "\\SDEC{"^toStringTerm(t)^"}{"^
				       toStringTermList(tl)^"}{"^
				       toStringTermList(xl)^"}{"^
				       toStringTerm(t0)^"}"^
				       "\\ATORIG{"^toStringSymb(c)^"}{\\{"^
				       toStringSymbList(cl)^"\\}}\\IN\\\\\n"^
				       toStringProc(P)

      | toStringProc(ADEC(t,tl,xl,t0,c,cl,_,P)) 
	                             = "\\ADEC{"^toStringTerm(t)^"}{"^
				       toStringTermList(tl)^"}{"^
				       toStringTermList(xl)^"}{"^
				       toStringTerm(t0)^"}"^
				       "\\ATORIG{"^toStringSymb(c)^"}{\\{"^
				       toStringSymbList(cl)^"\\}}\\IN\\\\\n"^
				       toStringProc(P)

      | toStringProc(PAR(P1,P2))     = toStringProc(P1)^"\\PAR\\\\\n"^toStringProc(P2)

      | toStringProc(BANG(P))        = "! "^toStringProc(P)

      | toStringProc(NEW(n,P))       = "\\NEW{"^toStringTerm(n)^"} "^toStringProc(P)

      | toStringProc(ANEW(n,P))      = "\\ANEW{"^toStringTerm(n)^"}"^toStringProc(P)

      | toStringProc(NIL)            = "\\NIL\\\\\n"			      

in
    fun toString P = toStringProc(P)

    fun toFile file P = let val fh = TextIO.openOut(file)
			    in
				(TextIO.output(fh, toStringProc(P));
				 TextIO.closeOut(fh))
			    end
end
end