(*****************************************************************************)
(* Data structure and auxiliary functions for Meta Labelled Lysa (MLLysa)    *)
(* Basis for meta level analysis                                             *)
(*                                                                           *)
(*                                 Time-stamp: <27-01-2005 Mikael Buchholtz> *)
(*****************************************************************************)
(* Datatype for Meta Labelled LySa is defined below.                         *)
(*                                                                           *)
(* NOTE: as for LySa the variables and names in input, decryption, and       *)
(* restriction are represented as terms (which means that they are also      *)
(* labelled)                                                                 *)
(*****************************************************************************)


structure MLLysa =
struct

type Index    = string
type Name     = string * Index list
type Var      = string * Index list
type CP       = string * Index list (* Crypto-points *)
type Lab      = string * Index list (* Term labels *)

datatype ISet = SID   of string 
	      | SET   of Index list
	      | UNION of ISet * ISet

datatype Term = NAME  of Name * Lab
	      | NAMEP of Name * Lab
	      | NAMEM of Name * Lab
	      | VAR   of Var  * Lab
	      | SENC  of Term list * Term * CP * CP list * Lab * Lab 
	      | AENC  of Term list * Term * CP * CP list * Lab * Lab 

datatype Proc = LET   of ISet * ISet * Proc
	      | IPAR  of Index * ISet * Proc
	      | INEW  of (Index * ISet) list * Term * Proc
	      | IANEW of (Index * ISet) list * Term * Proc
	      | OUT   of Term list * Proc
 	      | INP   of Term list * Term list * Proc			 
	      | SDEC  of Term *  Term list * Term list * Term *
			 CP * CP list * Lab * Proc	       
	      | ADEC  of Term *  Term list * Term list * Term *
			 CP * CP list * Lab * Proc	       
	      | NEW   of Term * Proc
	      | ANEW  of Term * Proc
	      | BANG  of Proc
	      | PAR   of Proc * Proc
	      | NIL

val CPDY      = ("_CPDY",[])

val NOLAB     = ("",[])

val NATURAL1  = SET(["1"]) 
val NATURAL2  = SET(["1","2"])	 
val NATURAL3  = SET(["1","2","3"])
val NATURAL01 = SET(["0","1"])
val NATURAL02 = SET(["0","1","2"])
val NATURAL03 = SET(["0","1","2","3"])
val ZERO      = SET(["0"])	  

local 
   open Set
in

fun substIndexIndex (i,a) (i1)   = if not(i1=i) then i1 else a

fun substIndexCP (i,a) ((n,il))  = (n,List.map (substIndexIndex (i,a)) il)

fun substIndexLab (i,a) ((l,il)) = (l,List.map (substIndexIndex (i,a)) il)

fun substIndexTerm (i,a) b (NAME((n,il),l))  = 
    NAME((n,List.map (substIndexIndex (i,a)) il), substIndexLab (i,a) l)
  | substIndexTerm (i,a) b (NAMEP((n,il),l)) = 
    NAMEP((n,List.map (substIndexIndex (i,a)) il), substIndexLab (i,a) l)
  | substIndexTerm (i,a) b (NAMEM((n,il),l)) = 
    NAMEM((n,List.map (substIndexIndex (i,a)) il), substIndexLab (i,a) l)
  | substIndexTerm (i,a) b (VAR((n,il),l))   = 
    VAR((n,List.map (substIndexIndex (i,a)) il), substIndexLab (i,a) l)
  | substIndexTerm (i,a) b (SENC(el,e0,c,cl,ls,l)) =     
    SENC(List.map (substIndexTerm (i,a) b) el, 
	substIndexTerm (i,a) b e0,
	substIndexCP (i,a) c,			      
	(List.map (substIndexCP (i,a)) cl)@(if (a=b) then [CPDY] else []),
	substIndexLab (i,a) ls,substIndexLab (i,a) l)
  | substIndexTerm (i,a) b (AENC(el,e0,c,cl,ls,l)) = 
    AENC(List.map (substIndexTerm (i,a) b) el, 
	 substIndexTerm (i,a) b e0,
	 substIndexCP (i,a) c,
	 List.map (substIndexCP (i,a)) cl@(if (a=b) then [CPDY] else []),
	 substIndexLab (i,a) ls,substIndexLab (i,a) l)

fun substIndexProc (i,a) b (LET(x,s,p))   = LET(x,s,substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (IPAR(i1,s,p)) = if i1=i then IPAR(i1,s,p)
					            else IPAR(i1,s,substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (INEW(al,n,p)) = INEW(al,n,substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (IANEW(al,n,p))= IANEW(al,n,substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (OUT(el,p))    = OUT(List.map (substIndexTerm (i,a) b) el,
						substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (INP(el,xl,p)) = INP(List.map (substIndexTerm (i,a) b) el,
						List.map (substIndexTerm (i,a) b) xl,
						substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (SDEC(e,el,xl,e0,c,cl,ls,p)) =
                                            SDEC(substIndexTerm (i,a) b e,
					         List.map (substIndexTerm (i,a) b) el,
					         List.map (substIndexTerm (i,a) b) xl,
						 substIndexTerm (i,a) b e0,
						 substIndexCP (i,a) c,
						 (List.map (substIndexCP (i,a)) cl)@
						 (if (a=b) then [CPDY] else []),
					         substIndexLab (i,a) ls,
					         substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (ADEC(e,el,xl,e0,c,cl,ls,p)) =
                                            ADEC(substIndexTerm (i,a) b e,
						 List.map (substIndexTerm (i,a) b) el,
						 List.map (substIndexTerm (i,a) b) xl,
						 substIndexTerm (i,a) b e0,
						 substIndexCP (i,a) c,
						 (List.map (substIndexCP (i,a)) cl)@
						 (if (a=b) then [CPDY] else []),
						 substIndexLab (i,a) ls,
						 substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (NEW(n,p))     = NEW(substIndexTerm (i,a) b n, 
						substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (ANEW(n,p))    = ANEW(substIndexTerm (i,a) b n, 
						 substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (BANG(p))      = BANG(substIndexProc (i,a) b p)
  | substIndexProc (i,a) b (PAR(p1,p2))   = PAR(substIndexProc (i,a) b p1, 
						substIndexProc (i,a) b p2)
  | substIndexProc (i,a) b (NIL)          = NIL


(*****************************************************************************)
(* namesProc : MLProc -> MLName list                                         *)
(* Returns the names in a process                                            *)
(* Labels at names are ignored                                               *)
(*****************************************************************************)

fun elm(SID(x)) = [x]
  | elm(SET(s)) = s
  | elm(UNION(s1,s2)) = (elm s1)@(elm s2)

fun namesTerm(NAME(n,_))    = [NAME(n,NOLAB)]
  | namesTerm(NAMEP(n,_))   = [NAMEP(n,NOLAB)]
  | namesTerm(NAMEM(n,_))   = [NAMEM(n,NOLAB)]
  | namesTerm(VAR(_,_))     = []
  | namesTerm(SENC(el,e0,c,cl,ls,l)) = unionList(List.map namesTerm (e0::el))
  | namesTerm(AENC(el,e0,c,cl,ls,l)) = unionList(List.map namesTerm (e0::el))

		
(* Note: First case in genIndices is a hack to prevent substitution to
   nullify names with indices not being substituted *)

fun genIndices (il)    (x,[]) =  if (List.exists (fn y => y=x) il) then [] else [il]
    
  | genIndices (i::il) (x,s)  = 
    let val ls = genIndices (il) (x,s)	
    in
	List.concat(List.map 
			(fn e => 
			    (List.map (fn ls' => (substIndexIndex (x,e) i)::ls') ls)) s)
    end

  | genIndices ([])    (x,s) = [[]]


fun substSeqName (x,s) (NAME((n,il),_))  = 
    List.map (fn il' => NAME((n,il'),NOLAB)) (genIndices il (x,s))
    
  | substSeqName (x,s) (NAMEP((n,il),_)) = 
    List.map (fn il' => NAMEP((n,il'),NOLAB)) (genIndices il (x,s))

  | substSeqName (x,s) (NAMEM((n,il),_)) = 
    List.map (fn il' => NAMEM((n,il'),NOLAB)) (genIndices il (x,s))

  | substSeqName (x,s) (_)               = 
    raise Fail "Only names can appear here"


fun substSeqCP (x,s) ((c,il)) = 
    List.map (fn il' => (c,il')) (genIndices il (x,s))
 

fun substIndexName (i,a) (NAME((n,il),_))  = NAME((n,List.map (substIndexIndex (i,a)) il), NOLAB)
  | substIndexName (i,a) (NAMEP((n,il),_)) = NAMEP((n,List.map (substIndexIndex (i,a)) il), NOLAB)
  | substIndexName (i,a) (NAMEM((n,il),_)) = NAMEM((n,List.map (substIndexIndex (i,a)) il), NOLAB)
  | substIndexName (i,a) (_)               = raise Fail "substIndexName: name expected"


fun namesProc(LET(SID(x),s,p)) = unionList(List.map (substSeqName (x,elm s))(namesProc p))
					
  | namesProc(LET(_,_,_))      = raise Fail "let: indexing identifier - not set - expected"

  | namesProc(IPAR(i,s,p))     = unionList(List.map 
					       (fn n => (List.map (fn e => (substIndexName (i,e) n)) (elm s))) 
					       (namesProc p))
  | namesProc(INEW(al,n,p))    = namesProc p
  | namesProc(IANEW(al,n,p))   = namesProc p
  | namesProc(OUT(el,p))       = union(unionList(List.map namesTerm el), 
				       namesProc(p))  
  | namesProc(INP(el,xl,p))    = union(unionList(List.map namesTerm el),
				       namesProc(p))
  | namesProc(SDEC(e,el,xl,e0,c,cl,ls,p))  
                               = union(unionList (map namesTerm (e::e0::el)),
				       namesProc(p))
  | namesProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                               = union(unionList (map namesTerm (e::e0::el)),
				       namesProc(p))
  | namesProc(NEW(NAME(n,_),p)) 
                               = union([NAME(n,NOLAB)], namesProc(p))
  | namesProc(NEW(_,_))        = raise Fail "Restriction must restrict a name"
  | namesProc(ANEW(NAME(n,_),p)) 
                               = union([NAMEP(n,NOLAB), NAMEM(n,NOLAB)], namesProc(p))
  | namesProc(ANEW(_,_))       = raise Fail "Restriction must restrict a name"
  | namesProc(BANG(p))         = namesProc(p)
  | namesProc(PAR(p1,p2))      = union(namesProc(p1), namesProc(p2))
  | namesProc(NIL)             = []



(*****************************************************************************)
(* mfnProc: Proc -> Name list                                                *)
(* Returns the free names in a process                                       *)
(* Labels are ignored and are instead replaced with NOLAB                    *)
(*****************************************************************************)
 
(*
fun mfnTerm(NAME(n,_))         = [NAME(n,NOLAB)]
  | mfnTerm(NAMEP(n,_))        = [NAMEP(n,NOLAB)]
  | mfnTerm(NAMEM(n,_))        = [NAMEM(n,NOLAB)]
  | mfnTerm(VAR(_,_))          = []
  | mfnTerm(SENC(el,e0,c,cl,ls,l))  
                               = unionList(List.map mfnTerm (e0::el))
  | mfnTerm(AENC(el,e0,c,cl,ls,l)) 
                               = unionList(List.map mfnTerm (e0::el))
*)

fun mfnTerm(e) = namesTerm(e)

fun mfnProc(LET(SID(x),s,p))   = 
    unionList(List.map (substSeqName (x,elm s)) (mfnProc p))
					
  | mfnProc(LET(_,_,_))        = raise Fail "let: must declare indexing identifier"

  | mfnProc(IPAR(i,s,p))       = 
    unionList(List.map 
		  (fn n => (List.map (fn e => (substIndexName (i,e) n)) (elm s))) 
		  (mfnProc p))


  | mfnProc(INEW(al,NAME((n,il),_),p)) =
    let val rn =
	    (List.foldl (fn ((i,s),nl) => 
			   (List.concat (List.map 
					     (fn e => List.map 
							   (fn n' => substIndexName (i,e) n') 
							   nl)
					     (elm s))))
			   
			[NAME((n,il),NOLAB)] al)

    in    
       difference(rn, mfnProc(p))	
    end

  | mfnProc(INEW(_,_,_))       = raise Fail "Restriction must restrict a name"
 

  | mfnProc(IANEW(al,NAME((n,il),_),p))     = 
    let val rn =
	    (List.foldl (fn ((i,s),nl) => 
			   (List.concat (List.map 
					     (fn e => List.map 
							   (fn n' => substIndexName (i,e) n') 
							   nl)
					     (elm s))))
			   
			[NAMEP((n,il),NOLAB),NAMEM((n,il),NOLAB)] al)

    in    
       difference(rn, mfnProc(p))
    end


  | mfnProc(IANEW(_,_,_))      = raise Fail "Restriction must restrict a name"

  | mfnProc(OUT(el,p))         = union(unionList(List.map mfnTerm el),
				       mfnProc(p))
  | mfnProc(INP(el,xl,p))      = union(unionList(List.map mfnTerm el),
				       mfnProc(p))				
  | mfnProc(SDEC(e,el,xl,e0,c,cl,ls,p)) 
                              = union(unionList(List.map mfnTerm (e::e0::el)),
				      mfnProc(p))
  | mfnProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                              = union(unionList(List.map mfnTerm (e::e0::el)),
				      mfnProc(p))
  | mfnProc(NEW(NAME(n,_),p))  = difference([NAME(n,NOLAB)], mfnProc(p))
  | mfnProc(NEW(_,_))          = raise Fail "Restriction must restrict a name"
  | mfnProc(ANEW(NAME(n,_),p)) = difference([NAMEP(n,NOLAB),NAMEM(n,NOLAB)], mfnProc(p))
  | mfnProc(ANEW(_,_))         = raise Fail "Restriction must restrict a name"
  | mfnProc(PAR(p1,p2))        = union(mfnProc(p1),mfnProc(p2))  
  | mfnProc(BANG(p))           = mfnProc(p)
  | mfnProc(NIL)               = []
 


(*****************************************************************************)
(* sencArityProc :  Proc -> int list                                         *)
(*      Finds arities of symmetric encrypted terms in the process.           *)
(*****************************************************************************)

fun sencArityTerm(NAME(_,_))           = []
  | sencArityTerm(NAMEP(_,_))          = []
  | sencArityTerm(NAMEM(_,_))          = []
  | sencArityTerm(VAR(_,_))            = []
  | sencArityTerm(SENC(el,e0,_,_,_,_)) = union([List.length el],
					       unionList(List.map sencArityTerm (e0::el)))
  | sencArityTerm(AENC(el,e0,_,_,_,_)) = unionList(List.map sencArityTerm (e0::el))			 


fun sencArityProc(LET(_,_,p))          = sencArityProc(p) 
  | sencArityProc(IPAR(_,_,p))         = sencArityProc(p) 
  | sencArityProc(INEW(_,_,p))         = sencArityProc(p) 
  | sencArityProc(IANEW(_,_,p))        = sencArityProc(p) 
  | sencArityProc(OUT(el,p))           = union(unionList(List.map sencArityTerm el),
					       sencArityProc(p))
  | sencArityProc(INP(el,_,p))         = union(unionList(List.map sencArityTerm el),
					       sencArityProc(p))
  | sencArityProc(SDEC(e,el,xl,e0,_,_,_,p))
                                       = union([List.length(el) + List.length(xl)],
					 union(unionList(List.map sencArityTerm (e::e0::el)),
					       sencArityProc(p)))
  | sencArityProc(ADEC(e,el,xl,e0,_,_,_,p)) 
                                       = union(unionList(List.map sencArityTerm (e::e0::el)),
					      sencArityProc(p))
  | sencArityProc(NEW(_,p))            = sencArityProc(p)
  | sencArityProc(ANEW(_,p))           = sencArityProc(p)
  | sencArityProc(BANG(p))             = sencArityProc(p)
  | sencArityProc(PAR(p1,p2))          = union(sencArityProc(p1),
					       sencArityProc(p2))
  | sencArityProc(NIL)                 = []



(*****************************************************************************)
(* aencArityProc :  Proc -> int list                                         *)
(*      Finds arities of asymmetric encrypted terms in the process.          *)
(*****************************************************************************)

fun aencArityTerm(NAME(_,_))           = []
  | aencArityTerm(NAMEP(_,_))          = []
  | aencArityTerm(NAMEM(_,_))          = []
  | aencArityTerm(VAR(_,_))            = []
  | aencArityTerm(SENC(el,e0,_,_,_,_)) = unionList(List.map aencArityTerm (e0::el))
  | aencArityTerm(AENC(el,e0,_,_,_,_)) = union([List.length el],
					       unionList(List.map aencArityTerm (e0::el)))


fun aencArityProc(LET(_,_,p))          = aencArityProc(p) 
  | aencArityProc(IPAR(_,_,p))         = aencArityProc(p) 
  | aencArityProc(INEW(_,_,p))         = aencArityProc(p) 
  | aencArityProc(IANEW(_,_,p))        = aencArityProc(p)  
  | aencArityProc(OUT(el,p))           = union(unionList(List.map aencArityTerm el),
					       aencArityProc(p))
  | aencArityProc(INP(el,_,p))         = union(unionList(List.map aencArityTerm el),
					       aencArityProc(p))
  | aencArityProc(SDEC(e,el,xl,e0,_,_,_,p))
                                       = union(unionList(List.map aencArityTerm (e::e0::el)),
					       aencArityProc(p))
  | aencArityProc(ADEC(e,el,xl,e0,_,_,_,p)) 
                                       = union([List.length(el) + List.length(xl)],
                                         union(unionList(List.map aencArityTerm (e::e0::el)),
					       aencArityProc(p))) 
  | aencArityProc(NEW(_,p))            = aencArityProc(p)
  | aencArityProc(ANEW(_,p))           = aencArityProc(p)
  | aencArityProc(BANG(p))             = aencArityProc(p)
  | aencArityProc(PAR(p1,p2))          = union(aencArityProc(p1),
					       aencArityProc(p2))
  | aencArityProc(NIL)                 = []
 


(*****************************************************************************)
(* comArityProc :  Proc -> int list                                        *)
(*      Finds arities of messages send by the process.                       *)
(*****************************************************************************)

fun comArityProc(LET(_,_,p))           = comArityProc(p) 
  | comArityProc(IPAR(_,_,p))          = comArityProc(p)
  | comArityProc(INEW(_,_,p))          = comArityProc(p)
  | comArityProc(IANEW(_,_,p))         = comArityProc(p)
  | comArityProc(OUT(el,p))            = union([(List.length el)],
						 comArityProc(p))
  | comArityProc(INP(el,xl,p))         = union([(List.length el +
						   List.length xl)],
						 comArityProc(p))
  | comArityProc(SDEC(_,_,_,_,_,_,_,p))= comArityProc(p)
  | comArityProc(ADEC(_,_,_,_,_,_,_,p))= comArityProc(p)
  | comArityProc(PAR(p1,p2))           = union(comArityProc(p1),
					       comArityProc(p2))
  | comArityProc(NEW(_,p))             = comArityProc(p)
  | comArityProc(ANEW(_,p))            = comArityProc(p)
  | comArityProc(BANG(p))              = comArityProc(p)
  | comArityProc(NIL)                  = []



(*****************************************************************************)
(* cryptoProc: Proc -> CP list                                               *)
(* Returns the  crypto-points in a process                                   *)
(*****************************************************************************)

fun cryptoTerm(NAME(_,_))         = []
  | cryptoTerm(NAMEP(_,_))        = []
  | cryptoTerm(NAMEM(_,_))        = []
  | cryptoTerm(VAR(_,_))          = []
  | cryptoTerm(SENC(el,e0,c,cl,ls,l)) 
                                  = union(c::cl,unionList(List.map cryptoTerm (e0::el)))
  | cryptoTerm(AENC(el,e0,c,cl,ls,l)) 
                                  = union(c::cl,unionList(List.map cryptoTerm (e0::el)))

fun cryptoProc(LET(SID(x),s,p))   = unionList(List.map (substSeqCP (x,elm s))(cryptoProc p))
  | cryptoProc(LET(_,_,_))        = raise Fail "let: indexing identifier - not set - expected"
  | cryptoProc(IPAR(i,s,p))       = unionList(List.map 
					       (fn n => (List.map (fn e => (substIndexCP (i,e) n)) (elm s))) 
					       (cryptoProc p))
  | cryptoProc(INEW(_,_,p))       = cryptoProc(p)
  | cryptoProc(IANEW(_,_,p))      = cryptoProc(p)
  | cryptoProc(OUT(el,p))         = union(unionList(List.map cryptoTerm el),
					  cryptoProc(p))
  | cryptoProc(INP(el,xl,p))      = union(unionList(List.map cryptoTerm el),
					  cryptoProc(p))
  | cryptoProc(SDEC(e,el,xl,e0,c,cl,ls,p)) 
                                  = union(c::cl,			    
			            union(unionList(List.map cryptoTerm (e::e0::el)),
					  cryptoProc(p)))
  | cryptoProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                                  = union(c::cl,			    
			            union(unionList(List.map cryptoTerm (e::e0::el)),
					  cryptoProc(p)))
  | cryptoProc(PAR(p1,p2))        = union(cryptoProc(p1),cryptoProc(p2))
  | cryptoProc(NEW(_,p))          = cryptoProc(p)
  | cryptoProc(ANEW(_,p))         = cryptoProc(p)
  | cryptoProc(BANG(p))           = cryptoProc(p)  
  | cryptoProc(NIL)               = []

end
end